//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

//---------------------------------------------------------------------------
//
//  TBDEUtilComp is a very simple conversion of the Borland TUtil32 program
//  to BCB4 component format.
//  (I think this should work in BCB3 and BCB5 as well)
//
//  This component functions exactly as the TUtil32 program since its code
//  is identical except for the necessary modifications for component format.
//  The only difference is that the resulting new files created will be in the
//  same directory as the table you are performing the operation on (as opposed
//  to the current directory of the program using the component).
//
//  Simply assign the full path to the Paradox table to rebuild or verify
//  to the TablePath property. (Make sure the table exists since this property
//  will reject assignment of non-functioning paths)
//
//  Then call VerifyTable() for verification,
//  and RebuildTable() for rebuilding. (both return boolean)
//
//  Examine the StatusMsg property to see what the result of the operation was.
//
//  Examine ErrorMsg to see the error message in the event of failure.
//
//  Examine Processing to see if the operation is still in progress.
//
//  I was going to provide support for the progress events which TUtil32 had,
//  but I ran into trouble trying to do it..  Basically it would GPF either
//  the tutil32.dll or kernel32.dll whenever I called the assigned TNotifyEvent
//  function (during the GenProgress callback) for no aparrent reason..  Anyways
//  in the end I gave up and cut out all my progress events and commented
//  out the progress stuff from the original TUtil32 source code. Sorry.
//
//  IMPORTANT:
//    You must add tutil32.lib to any application's project manager in which you
//    use this component and after that your program will require tutil32.dll to
//    execute.  Make sure you use the correct version of tutil32.dll for the
//    version of the BDE which you are using.
//    See http://www.borland.com/devsupport/bde/utilities.html
//
//
//  DISCLAIMER:
//    Unfortunately this 100% free component is just as unsupported by me as the
//    unsupported Borland source code on which it is based.  Please do not e-mail
//    me with usage questions, just read through this extremely short source code.
//    If you like you may e-mail me suggestions and/or your modifications/enhancements.
//    (Or even your thanks if you like, hehe.)
//
//                              Jonathan W. McCallister
//                              cecil@ImagingA.com
//                              4/30/2000
//---------------------------------------------------------------------------

#include "BDEUtilComp.h"


// Includes
#include <mem.h>
#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include <stdlib.h>

#pragma package(smart_init)

// Global pointer for use in callback functions
TBDEUtilComp* BDEUtilComp = NULL;
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(TBDEUtilComp *)
{
  new TBDEUtilComp(NULL);
}
//---------------------------------------------------------------------------
__fastcall TBDEUtilComp::TBDEUtilComp(TComponent* Owner)
        : TComponent(Owner)
{
  BDEUtilComp = this;
  ProcessingTable = false;
  BDEUtil = new TBDEUtil();
}
//---------------------------------------------------------------------------
__fastcall TBDEUtilComp::~TBDEUtilComp()
{
  delete BDEUtil;
}
//---------------------------------------------------------------------------
void __fastcall TBDEUtilComp::SetTablePath(String sTableFile)
{
  bool Success = true;

  if ( !ProcessingTable )
  {
    if ( !sTableFile.IsEmpty() )
      Success = FileExists(sTableFile);

    if ( Success )
      TPath = sTableFile;
  }
}
//---------------------------------------------------------------------------
bool TBDEUtilComp::VerifyTable()
{
  if ( TPath.IsEmpty() )
  {
    StatusStr = "Verification Failed: No table is specified yet.";
    ErrorStr = "No table specified";
    return false;
  }

  bool Success = true;

  // [Slightly] Modified TUtil32 application code
  int Msg;
  AnsiString Table;
  unsigned short far* temp = new unsigned short(BDEUtil->vhTSes);
  Table = TPath;
  String DirStr = TPath.SubString(0, TPath.LastDelimiter("\\"));

  TUCheck(TUExit(BDEUtil->vhTSes));
  TUCheck(TUInit(temp));
  if (BDEUtil->RegisterCallBack())
  {
    ProcessingTable = true;
    Screen->Cursor = crHourGlass;

    try
    {
      if (TUCheck(TUVerifyTable(BDEUtil->vhTSes, Table.c_str(), szPARADOX,
          (DirStr+"VERIFY.DB").c_str(), NULL, 0, (unsigned int*)&Msg)) == DBIERR_NONE)
      {
        switch(Msg)
        {
          case 0:
            StatusStr = "Verification Successful.  Table has no errors.";
            break;

          case 1:
            StatusStr = "Verification Successful.  Verification completed.";
            break;

          case 2:
            StatusStr = "Verification Successful.  Verification could not be completed.";
            break;

          case 3:
            StatusStr = "Verification Successful.  Table must be rebuilt manually.";
            break;

          case 4:
            StatusStr = "Verification Successful.  Table cannot be rebuilt.";
            break;
        }
      }
      else
      {
        Success = false;
        StatusStr = "Verification Failed!";
      }
    }
    catch (Exception &e)
    {
      Success = false;
      StatusStr = "Verification Exception: "+e.Message;
    }

    ProcessingTable = false;

    BDEUtil->UnRegisterCallBack();
    Screen->Cursor = crDefault;
  }

  delete temp;

  return Success;
}
//---------------------------------------------------------------------------
bool TBDEUtilComp::RebuildTable()
{
  if ( TPath.IsEmpty() )
  {
    StatusStr = "Rebuild Failed: No table is specified yet.";
    ErrorStr = "No table specified";
    return false;
  }

  bool Success = true;

  // [Slightly] Modified TUtil32 application code
  unsigned short iFld, iIdx, iSec, iVal, iRI, iOptP, iOptD;
  AnsiString szTable;
  DBIResult rslt;
  CRTblDesc TblDesc;
  AnsiString Backup;
  int Msg;
  unsigned short far* temp = new unsigned short(BDEUtil->vhTSes);

  TUCheck(TUExit(BDEUtil->vhTSes));
  TUCheck(TUInit(temp));
  szTable = TPath;
  String DirStr = TPath.SubString(0, TPath.LastDelimiter("\\"));
  if (BDEUtil->RegisterCallBack())
  {
    ProcessingTable = true;
    Screen->Cursor = crHourGlass;

    try
    {
      if (TUCheck(TUVerifyTable(BDEUtil->vhTSes, szTable.c_str(),
        szPARADOX, (DirStr+"VERIFY.DB").c_str(),NULL, 0, (unsigned int*)&Msg)) == DBIERR_NONE)
      {
        if (TUCheck(TUGetCRTblDescCount(BDEUtil->vhTSes, szTable.c_str(), &iFld,
            &iIdx, &iSec, &iVal, &iRI, &iOptP, &iOptD)) == DBIERR_NONE)
        {
          memset(&TblDesc, 0, sizeof(CRTblDesc));
          strcpy(TblDesc.szTblName, szTable.c_str());
          strcpy(TblDesc.szTblType, szPARADOX);
          strcpy(TblDesc.szErrTblName, "rebuild.db");

          TblDesc.iFldCount = iFld;
          TblDesc.pfldDesc = (pFLDDesc) malloc (iFld * sizeof(FLDDesc));

          TblDesc.iIdxCount = iIdx;
          TblDesc.pidxDesc = (pIDXDesc) malloc (iIdx * sizeof (IDXDesc));

          TblDesc.iSecRecCount = iSec;
          TblDesc.psecDesc = (pSECDesc) malloc (iSec * sizeof(SECDesc));

          TblDesc.iValChkCount = iVal;
          TblDesc.pvchkDesc = (pVCHKDesc) malloc (iVal * sizeof(VCHKDesc));

          TblDesc.iRintCount = iRI;
          TblDesc.printDesc = (pRINTDesc) malloc (iOptD * sizeof(RINTDesc));

          TblDesc.iOptParams = iOptP;
          TblDesc.pfldOptParams = (pFLDDesc) malloc (iOptP * sizeof(FLDDesc));

          TblDesc.pOptData = (pBYTE) malloc(iOptD * DBIMAXSCFLDLEN);

          if (TUCheck(TUFillCRTblDesc(BDEUtil->vhTSes, &TblDesc, szTable.c_str(),
                      NULL)) == DBIERR_NONE)
          {
            Backup = DirStr+"Backup.db";

            if (TUCheck(TURebuildTable(BDEUtil->vhTSes, szTable.c_str(), szPARADOX,
                Backup.c_str(), (DirStr+"KEYVIOL.DB").c_str(), (DirStr+"PROBLEM.DB").c_str(),
                &TblDesc)) == DBIERR_NONE)
            {
              StatusStr = "Rebuild was successful";
            }
            else
            {
              Success = false;
              StatusStr = "Rebuild was not successful";
              ErrorStr = "TURebuildTable() Failed";
            }
          }
          else
          {
            Success = false;
            StatusStr = "Rebuild Failed.";
            ErrorStr = "Error Filling table structure.";
          }

          free(TblDesc.pfldDesc);
          free(TblDesc.pidxDesc);
          free(TblDesc.psecDesc);
          free(TblDesc.pvchkDesc);
          free(TblDesc.printDesc);
          free(TblDesc.pfldOptParams);
          free(TblDesc.pOptData);
        }
        else
        {
          Success = false;
          StatusStr = "Rebuild Failed.";
          ErrorStr = "Error getting description count.";
        }
      }
      else
      {
        Success = false;
        StatusStr = "Rebuild: Verification Failed!";
        ErrorStr = "Could not rebuild table because verification failed.";
      }
    }
    catch (Exception &e)
    {
      if ( TblDesc.pfldDesc )
        free(TblDesc.pfldDesc);
      if ( TblDesc.pidxDesc )
        free(TblDesc.pidxDesc);
      if ( TblDesc.psecDesc )
        free(TblDesc.psecDesc);
      if ( TblDesc.pvchkDesc )
        free(TblDesc.pvchkDesc);
      if ( TblDesc.printDesc )
        free(TblDesc.printDesc);
      if ( TblDesc.pfldOptParams )
        free(TblDesc.pfldOptParams);
      if ( TblDesc.pOptData )
        free(TblDesc.pOptData);

      Success = false;
      StatusStr = "Rebuild: Exception was thrown.";
      ErrorStr = "Rebuild: Exception: "+e.Message;
    }

    ProcessingTable = false;
    BDEUtil->UnRegisterCallBack();
    Screen->Cursor = crDefault;
  }

  delete temp;

  return Success;
}
//---------------------------------------------------------------------------
namespace Bdeutilcomp
{
  void __fastcall PACKAGE Register()
  {
    TComponentClass classes[1] = {__classid(TBDEUtilComp)};
    RegisterComponents("Data Access", classes, 0);
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

// [Slightly] Modified original TUtil32 sourcecode.

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
DBIResult CheckDbiError(DBIResult ErrorValue, char* FileName, unsigned long LineNo)
{
  // This function is used to check all DbiFunctions
  //
  char        dbi_status[DBIMAXMSGLEN * 5] = {'\0'}; // Error String
  DBIMSG      dbi_string = {'\0'};
  DBIErrInfo  ErrInfo;    // Contains information about the error

  if (ErrorValue != DBIERR_NONE)
  {
    // Note - make certain to call DbiGetErrorInfo() right after
    // the error as it will only give information about the most
    // recent error.
    DbiGetErrorInfo(TRUE, ErrInfo);

    // Check to see if DbiGetErrorInfo is reporting information
    //   on the same error that you have. (Done in case DbiGetErrorInfo
    //   was not called immediately after the function which failed)
    if (ErrInfo.iError == ErrorValue)
    {
      wsprintf(dbi_status, "  ERROR %i - Cat:Code = [%xh:%xh]\r\n    %s",
               ErrorValue, ErrCat(ErrInfo.iError), ErrCode(ErrInfo.iError),
               ErrInfo.szErrCode);

      // Need to check how much information was provided -
      //   different errors return different amounts of information.
      if (strcmp(ErrInfo.szContext[1], ""))
      {
        wsprintf(dbi_status, "%s\r\n    %s", dbi_status,
                 ErrInfo.szContext[1]);
      }
      if (strcmp(ErrInfo.szContext[2], ""))
      {
        wsprintf(dbi_status, "%s\r\n    %s", dbi_status,
                 ErrInfo.szContext[2]);
      }
      if (strcmp(ErrInfo.szContext[3], ""))
      {
        wsprintf(dbi_status, "%s\r\n    %s", dbi_status,
                 ErrInfo.szContext[3]);
      }
      if (strcmp(ErrInfo.szContext[4], ""))
      {
        wsprintf(dbi_status, "%s\r\n    %s", dbi_status,
                 ErrInfo.szContext[4]);
      }
    }
    else
    {
      DbiGetErrorString(ErrorValue, dbi_string);
      wsprintf(dbi_status, "  ERROR %i - Cat:Code = [%xh:%xh]\r\n    %s",
               ErrorValue, ErrCat(ErrorValue), ErrCode(ErrorValue),
               dbi_string);
    }

    // Add the file name and line number to the error string
    wsprintf(dbi_status, "%s\r\n    FileName: '%s'  Line Number: %i",
             dbi_status, FileName, LineNo);

    if ( BDEUtilComp )
      BDEUtilComp->ErrorMsg = String(dbi_status);

/*
    // Display error in snipit and in a MessageBox
    MessageBeep(MB_ICONEXCLAMATION);
    MessageBox(GetFocus(), dbi_status, "TUtility -- BDE Error",
               MB_OK | MB_ICONEXCLAMATION);
*/
  }

  return ErrorValue;
}
//---------------------------------------------------------------------------
DBIResult CheckTUError(DBIResult ErrorValue, char * FileName, unsigned long LineNo)
{
  // Check all TUTility functions
  CHAR dbi_status[DBIMAXMSGLEN * 2], dbi_string[DBIMAXMSGLEN];

  if (ErrorValue != DBIERR_NONE)
  {
    TUGetErrorString(ErrorValue, dbi_string);
    wsprintf(dbi_status, "%s\r\n    FileName: '%s'  Line Number: %i",
             dbi_string, FileName, LineNo);

    if ( BDEUtilComp )
      BDEUtilComp->ErrorMsg = String(dbi_status);
/*
    MessageBeep(MB_ICONEXCLAMATION);
    MessageBox(GetFocus(), dbi_status, "TUtility -- TU Error",
               MB_OK | MB_ICONEXCLAMATION);
*/
  }

  return ErrorValue;
}
//---------------------------------------------------------------------------
//  Updates the progress bar with the Status of the verify or rebuild
//
//---------------------------------------------------------------------------
CBRType __stdcall GenProgressCallBack(CBType ecbType, int Data, void* pcbInfo)
{
  TUVerifyCallBack* CBInfo;

  if (ecbType == cbGENPROGRESS)
  {
    CBInfo = (TUVerifyCallBack far*)pcbInfo;

    if ( BDEUtilComp )
    {
      switch (CBInfo->Process)
      {
        case TUVerifyHeader:
//          MainDialog->PBHeader->Position = CBInfo->PercentDone;
          break;

        case TUVerifyIndex:
//          MainDialog->PBIndexes->Position = CBInfo->PercentDone;
          break;

        case TUVerifyData:
//          MainDialog->PBData->Position = CBInfo->PercentDone;
          break;

        case TURebuild:
//          MainDialog->PBRebuild->Position = CBInfo->PercentDone;
          break;
      }
    }
  }

  return cbrUSEDEF;
}
//---------------------------------------------------------------------------
// Class implementation of the TBDEUtil class
//
//---------------------------------------------------------------------------
TBDEUtil::TBDEUtil()
{
  // Default Constructor...
  //
  hDb = 0;
  TUCheck(TUInit(&vhTSes));
}
//---------------------------------------------------------------------------
TBDEUtil::~TBDEUtil()
{
  // Default destructor...
  //
  TUExit(vhTSes);
}
//---------------------------------------------------------------------------
bool TBDEUtil::GetTCursorProps(char* szTable)
{
  return (TUCheck(TUFillCURProps(vhTSes, szTable, &tuProps)) == DBIERR_NONE);
}
//---------------------------------------------------------------------------
bool TBDEUtil::RegisterCallBack()
{
//  CBRType(__stdcall *CallBackfp)(CBType, long, void*);
//  CallBackfp = &GenProgressCallBack;

/*
  char CallbackData[100 * sizeof(char)];

  DBIResult rslt = Check(DbiRegisterCallBack(NULL, cbGENPROGRESS, (unsigned int)CallbackData,
                         sizeof(TUVerifyCallBack), &cbInfo, GenProgressCallBack));
  return (rslt == DBIERR_NONE);
*/
  return true;
}
//---------------------------------------------------------------------------
bool TBDEUtil::UnRegisterCallBack()
{
/*
  DBIResult rslt = Check(DbiRegisterCallBack(NULL, cbGENPROGRESS, (unsigned int)NULL,
                         sizeof(TUVerifyCallBack), &cbInfo, GenProgressCallBack));
  return (rslt == DBIERR_NONE);
*/

  return true;
}
//---------------------------------------------------------------------------
